/*****************************************************************************/
/*
                                FileDot.c

Miscellaneous file processing module.
(We might end up stuffing all sort of miscellaneous file-type functionality
into here!)


FILE.URL
--------
This file type parallels the Microsoft 'file.URL' files that contain Web URLs
as a type of symbolic or soft link into the Internet.  These files are enabled
by creating an HTTPD$CONFIG content-type mapping a file extension (e.g. .URL)
to the content-type "x-internal/url".  This results in FileDotUrlBegin() being
actiavted by REQUEST.C module when a file of type .URL is accessed.  The file
may contain a Microsoft style "URL=scheme://the.host.name/the/path/" URL, or a
plain Web URL such as "scheme://the.host.name/the/path/".  Both cause a
redirection to the specified URL.


VERSION HISTORY
---------------
01-JAN-2001  MGD  new for HTTPd version 7.2
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>

/* VMS related header files */
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"

#define WASD_MODULE "DOTFILE"

/**********/
/* macros */
/**********/

/* if not in the first 2048 bytes of the file then quit */
#define FILE_DOT_URL_QUIT_AFTER 2048

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  WatchEnabled;
extern char  SoftwareID[];
extern ACCOUNTING_STRUCT  *AccountingPtr;
extern MSG_STRUCT  Msgs;

/*****************************************************************************/
/*
Using the configuration file (contents supplied by a preceding call to
FileBegin()) interpret the mappings.
*/

FileDotUrlBegin (REQUEST_STRUCT *rqptr)

{
   int  Length;
   char  *cptr, *sptr, *zptr;
   char  FileName [ODS_MAX_FILE_NAME_LENGTH+1];
   FILE_CONTENT  *fcptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FileDotUrlBegin()\n");

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++AccountingPtr->DoDotUrlCount;

   /* take control of the file contents structure */
   fcptr = rqptr->FileContentPtr;
   rqptr->FileContentPtr = NULL;

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_RESPONSE))
      WatchThis (rqptr, FI_LI, WATCH_RESPONSE, "URL !AZ !UL bytes",
                 fcptr->FileName, fcptr->ContentLength);

   /* get the start of the file contents */
   for (cptr = sptr = fcptr->ContentPtr; *cptr; cptr++)
   {
      if (toupper(cptr[0]) == 'U')
      {
         if (cptr > sptr + FILE_DOT_URL_QUIT_AFTER) break;
         if (toupper(cptr[1]) != 'R') continue;
         if (toupper(cptr[2]) != 'L') continue;
         if (cptr[3] != '=') continue;
         /* hmmm, looks like a "URL=scheme://the.host.name/the/path" */
         sptr = (cptr += 4);
         while (*cptr && !isspace(*cptr)) cptr++;
         rqptr->rqResponse.LocationPtr = VmGetHeap (rqptr, cptr-sptr+1);
         cptr = rqptr->rqResponse.LocationPtr;
         while (*sptr && !isspace(*sptr)) *cptr++ = *sptr++;
         *cptr = '\0'; 
         break;
      }
      /* a possible Web URL (e.g. "scheme://the.host.name/the/path") */
      if (cptr[0] != ':') continue;
      if (cptr > sptr + FILE_DOT_URL_QUIT_AFTER) break;
      if (cptr[1] != '/') continue;
      if (cptr[2] != '/') continue;
      /* yup, looks like it, find the start of the scheme, looking backwards */
      if (cptr > fcptr->ContentPtr) cptr--; 
      while (cptr > fcptr->ContentPtr && isalpha(*cptr)) cptr--; 
      if (!isalpha(*cptr)) cptr++;
      sptr = cptr;
      /* find the end of the URL, then use it as a redirection location */
      while (*cptr && !isspace(*cptr)) cptr++;
      rqptr->rqResponse.LocationPtr = VmGetHeap (rqptr, cptr-sptr+1);
      cptr = rqptr->rqResponse.LocationPtr;
      while (*sptr && !isspace(*sptr)) *cptr++ = *sptr++;
      *cptr = '\0'; 
      break;
   }

   if (rqptr->rqResponse.LocationPtr != NULL)
   {
      /************/
      /* redirect */
      /************/

      if (Debug)
         fprintf (stdout, "LocationPtr |%s|\n", rqptr->rqResponse.LocationPtr);

      SysDclAst (fcptr->NextTaskFunction, rqptr);
      return;
   }

   /*****************************/
   /* output file as plain-text */
   /*****************************/

   HttpHeader (rqptr, 200, "text/plain", -1, NULL, NULL);

   NetWrite (rqptr, fcptr->NextTaskFunction,
             fcptr->ContentPtr, fcptr->ContentLength);
}

/*****************************************************************************/

